/*
 *  Created on 28.11.2003
 *
 */
package seco.semcocoon.transformation;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.Serializable;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.LinkedList;
import java.util.Map;

import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.sax.SAXTransformerFactory;
import javax.xml.transform.sax.TransformerHandler;
import javax.xml.transform.stream.StreamResult;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.parameters.Parameters;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.caching.CacheableProcessingComponent;
import org.apache.cocoon.environment.SourceResolver;
import org.apache.cocoon.transformation.ServiceableTransformer;
import org.apache.excalibur.source.SourceValidity;
import org.apache.excalibur.source.impl.validity.NOPValidity;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;
/**
 *@author     ssaarela
 *@created    16. joulukuuta 2003
 */
public class LoggingTransformer extends ServiceableTransformer implements Configurable, CacheableProcessingComponent {
    private boolean forward = false;
    private ByteArrayOutputStream logEntry = null;
    private TransformerHandler logHandler;
    private LogWriter logWriter = new LogWriter();
    private AttributesImpl nsAttributes = null;
    private boolean omitXmlDeclaration = true;
    private boolean record = false;
    private String startElement = null;
    private boolean stripRoot = true;
    /**
     *  Description of the Field
     */
    protected SAXTransformerFactory tfactory = (SAXTransformerFactory) TransformerFactory.newInstance();

	private String key;

	public Serializable getKey() {
		return key;
	}
	
	public SourceValidity getValidity() {
		return NOPValidity.SHARED_INSTANCE;
	}

    /**
     *  Description of the Method
     *
     *@param  ch                Description of the Parameter
     *@param  start             Description of the Parameter
     *@param  length            Description of the Parameter
     *@exception  SAXException  Description of the Exception
     */
    public void characters(char[] ch, int start, int length) throws SAXException {
        if (forward) {
            super.characters(ch, start, length);
        }
        if (record) {
            this.logHandler.characters(ch, start, length);
        }
    }


    /**
     *  Description of the Method
     *
     *@param  conf                        Description of the Parameter
     *@exception  ConfigurationException  Description of the Exception
     */
    public void configure(Configuration conf) throws ConfigurationException {
        Configuration tmp = conf.getChild("logDir", false);
        String dir = "logs";
        if (tmp != null) {
            dir = tmp.getValue();
            logWriter.setLogDir(dir);
        }
        tmp = conf.getChild("stripRoot", false);
        if (tmp != null) {
            stripRoot = new Boolean(tmp.getValue()).booleanValue();
        }
        tmp = conf.getChild("forward", false);
        if (tmp != null) {
            forward = new Boolean(tmp.getValue()).booleanValue();
        }
        this.key = dir+stripRoot+forward; 
        this.logWriter.start();
    }


    /**
     *  Description of the Method
     *
     *@exception  SAXException  Description of the Exception
     */
    public void endDocument() throws SAXException {
        if (forward) {
            super.endDocument();
        }
        this.logHandler.endDocument();
        this.logWriter.addLogEvent(this.logEntry.toByteArray());
    }


    /**
     *  Description of the Method
     *
     *@param  uri               Description of the Parameter
     *@param  localName         Description of the Parameter
     *@param  qName             Description of the Parameter
     *@exception  SAXException  Description of the Exception
     */
    public void endElement(String uri, String localName, String qName) throws SAXException {
        if (forward) {
            super.endElement(uri, localName, qName);
        }
        if (record && !(uri + localName).equals(this.startElement)) {
            this.logHandler.endElement(uri, localName, qName);
        } else {
            record = false;
        }
    }


    /**
     *  Description of the Method
     *
     *@param  prefix            Description of the Parameter
     *@exception  SAXException  Description of the Exception
     */
    public void endPrefixMapping(String prefix) throws SAXException {
        if (forward) {
            super.endPrefixMapping(prefix);
        }
        this.logHandler.endPrefixMapping(prefix);
    }



    /**
     *  Description of the Method
     *
     *@param  ch                Description of the Parameter
     *@param  start             Description of the Parameter
     *@param  length            Description of the Parameter
     *@exception  SAXException  Description of the Exception
     */
    public void ignorableWhitespace(char[] ch, int start, int length) throws SAXException {
        if (forward) {
            super.ignorableWhitespace(ch, start, length);
        }
        if (record) {
            this.logHandler.ignorableWhitespace(ch, start, length);
        }
    }


    /**
     *  Description of the Method
     *
     *@param  target            Description of the Parameter
     *@param  data              Description of the Parameter
     *@exception  SAXException  Description of the Exception
     */
    public void processingInstruction(String target, String data) throws SAXException {
        if (forward) {
            super.processingInstruction(target, data);
        }
        if (record) {
            this.logHandler.processingInstruction(target, data);
        }
    }


    /**
     *  Description of the Method
     */
    public void recycle() {
        super.recycle();
        this.logHandler = null;
        this.record = false;
        this.startElement = null;
        this.logEntry = null;
    }


    /**
     *  Sets the documentLocator attribute of the LoggingTransformer object
     *
     *@param  locator  The new documentLocator value
     */
    public void setDocumentLocator(Locator locator) {
        if (forward) {
            super.setDocumentLocator(locator);
        }
        if (record) {
            this.logHandler.setDocumentLocator(locator);
        }
    }


    /**
     *  Description of the Method
     *
     *@param  resolver                 Description of the Parameter
     *@param  objectModel              Description of the Parameter
     *@param  src                      Description of the Parameter
     *@param  parameters               Description of the Parameter
     *@exception  ProcessingException  Description of the Exception
     *@exception  SAXException         Description of the Exception
     *@exception  IOException          Description of the Exception
     */
    public void setup(SourceResolver resolver, Map objectModel, String src, Parameters parameters) throws ProcessingException, SAXException, IOException {
        super.setup(resolver, objectModel, src, parameters);
    }


    /**
     *  Description of the Method
     *
     *@param  name              Description of the Parameter
     *@exception  SAXException  Description of the Exception
     */
    public void skippedEntity(String name) throws SAXException {
        if (forward) {
            super.skippedEntity(name);
        }
        if (record) {
            this.logHandler.skippedEntity(name);
        }
    }


    /**
     *  Description of the Method
     *
     *@exception  SAXException  Description of the Exception
     */
    public void startDocument() throws SAXException {
        try {
            this.logEntry = new ByteArrayOutputStream();
            this.logHandler = tfactory.newTransformerHandler();
            Transformer transformer = this.logHandler.getTransformer();
            transformer.setOutputProperty(OutputKeys.METHOD, "xml");
            transformer.setOutputProperty(OutputKeys.ENCODING, "ISO-8859-1");
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.setOutputProperty("omit-xml-declaration", omitXmlDeclaration ? "yes" : "no");
            this.logHandler.setResult(new StreamResult(this.logEntry));
        } catch (TransformerConfigurationException e) {
            throw new SAXException(e);
        }
        if (forward) {
            super.startDocument();
        }
        this.logHandler.startDocument();
    }


    /**
     *  Description of the Method
     *
     *@param  uri               Description of the Parameter
     *@param  localName         Description of the Parameter
     *@param  qName             Description of the Parameter
     *@param  atts              Description of the Parameter
     *@exception  SAXException  Description of the Exception
     */
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        if (forward) {
            super.startElement(uri, localName, qName, atts);
        }
        if (!stripRoot) {
            record = true;
        }
        if (record) {
            if (nsAttributes != null) {
                atts = nsAttributes;
                nsAttributes = null;
            }
            this.logHandler.startElement(uri, localName, qName, atts);
        }
        if (!record) {
            record = true;
            startElement = uri + localName;
            nsAttributes = new AttributesImpl(atts);
        }
    }


    /**
     *  Description of the Method
     *
     *@param  prefix            Description of the Parameter
     *@param  uri               Description of the Parameter
     *@exception  SAXException  Description of the Exception
     */
    public void startPrefixMapping(String prefix, String uri) throws SAXException {
        if (forward) {
            super.startPrefixMapping(prefix, uri);
        }
        if (record) {
            this.logHandler.startPrefixMapping(prefix, uri);
        }
    }
}

/**
 *  Description of the Class
 *
 *@author     ssaarela
 *@created    17. joulukuuta 2003
 */
class LogWriter extends Thread {


    private DateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd");
    private Object lock = null;
    private File logDir = null;
    private LinkedList logEvents = new LinkedList();
    private File logFile = null;
    private FileOutputStream outputStream = null;
    private GregorianCalendar streamDate;


    /**
     *  Adds a feature to the LogEvent attribute of the LogWriter object
     *
     *@param  logEvent  The feature to be added to the LogEvent attribute
     */
    public synchronized void addLogEvent(byte[] logEvent) {
        this.logEvents.addLast(logEvent);
        notifyAll();
    }


    /**
     *  Gets the outputStream attribute of the LoggingTransformer class
     *
     *@return                  The outputStream value
     *@exception  IOException  Description of the Exception
     */
    private synchronized OutputStream getOutputStream() throws IOException {
        GregorianCalendar currentDate = new GregorianCalendar();
        if (outputStream == null
                 || !logFile.exists()
                 || currentDate.get(Calendar.DAY_OF_MONTH) != streamDate.get(Calendar.DAY_OF_MONTH)
                 || currentDate.get(Calendar.MONTH) != streamDate.get(Calendar.MONTH)) {
            if (outputStream != null) {
                outputStream.flush();
                outputStream.close();
            }
            streamDate = currentDate;
            logFile = new File(logDir, "MuseumFinland_" + dateFormat.format(streamDate.getTime()) + ".log");
            outputStream = new FileOutputStream(logFile, true);
        }
        return outputStream;
    }



    /**
     *  Main processing method for the LogWriter object
     */
    public synchronized void run() {
        while (!interrupted()) {
            try {
                while (this.logEvents.size() == 0) {
                    wait();
                }
            } catch (InterruptedException ie) {}
            while (this.logEvents.size() > 0) {
                synchronized (this.lock) {
                    try {
                        OutputStream out = getOutputStream();
                        out.write((byte[]) this.logEvents.removeFirst());
                        out.write("\n\n".getBytes());
                        out.flush();
                    } catch (IOException ioe) {
                        System.out.println("Logging failed!");
                        ioe.printStackTrace();
                    }
                }
            }
        }
    }


    /**
     *  Sets the logDir attribute of the LogWriter object
     *
     *@param  dir  The new logDir value
     */
    public synchronized void setLogDir(String dir) {
        logDir = new File(dir);
        if (!logDir.exists()) {
            logDir.mkdirs();
        }
        this.lock = logDir.getAbsoluteFile().toString().intern();
        System.out.println("Logging into " + this.lock);
    }
    
}

