/*
 *  Created on 28.11.2003
 *
 */
package seco.semcocoon.transformation;

import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.io.Serializable;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.parameters.ParameterException;
import org.apache.avalon.framework.parameters.Parameters;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.caching.CacheableProcessingComponent;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.SourceResolver;
import org.apache.excalibur.source.SourceValidity;
import org.xml.sax.SAXException;

import seco.util.Base64;
import sun.net.www.protocol.http.HttpURLConnection;

/**
 *@author     jiemakel
 *@created    4. joulukuuta 2003
 */
public class XMLHTTPServerTransformer extends AbstractSerializedXMLTransformer implements Configurable, CacheableProcessingComponent {

    /**
     *  Description of the Field
     */
    protected Configuration[] locations = null;
    /**
     *  Description of the Field
     */
    protected String passwordDefault = null;
    /**
     *  Description of the Field
     */
    protected String usernameDefault = null;
    
    protected HttpURLConnection connection = null;
    
    private static Map validityMap = new HashMap();

	/* (non-Javadoc)
	 * @see org.apache.cocoon.caching.CacheableProcessingComponent#getKey()
	 */
	public Serializable getKey() {
		StringBuffer locationString = new StringBuffer();
		for (int i=0;i<locations.length;i++) {
			try {
				locationString.append(locations[i].getValue());
			} catch (ConfigurationException e) {
			}
		}
		return parameters.getParameter("username","-")+parameters.getParameter("password","-")+parameters.getParameter("location","-")+usernameDefault+passwordDefault+locationString;
	}

	public SourceValidity getValidity() {
		ModifiableTimeStampValidity validity = (ModifiableTimeStampValidity)validityMap.get(getKey());
		if (validity == null) {
			validity= new ModifiableTimeStampValidity();
			validityMap.put(getKey(),validity);
		} 
		return validity;
	}

    /**
     *  Description of the Method
     *
     *@param  conf                        Description of the Parameter
     *@exception  ConfigurationException  Description of the Exception
     */
    public void configure(Configuration conf)
             throws ConfigurationException {
        locations = conf.getChildren("location");
        try {
            usernameDefault = conf.getChild("username").getValue();
            passwordDefault = conf.getChild("password").getValue();
        } catch (ConfigurationException e) {
            //e.printStackTrace();
            usernameDefault = null;
            passwordDefault = null;
        }
    }

	public void setup(SourceResolver resolver, Map objectModel, String src, Parameters par) throws ProcessingException, SAXException, IOException {
		super.setup(resolver,objectModel,src,par);
		Request request = ObjectModelHelper.getRequest(this.objectModel);
		connection = null;
		String username = null;
		String password = null;
		try {
			username = parameters.getParameter("username");
			password = parameters.getParameter("password");
		} catch (ParameterException e) {
			username = usernameDefault;
			password = passwordDefault;
		}
		try {
			connection = getConnection(new URL(parameters.getParameter("location")+"?"+request.getQueryString()), username, password);
			return;
		} catch (ParameterException e) {
		} catch (Exception e) {
			try {
				System.out.println("HTTPServerTransformer at " + parameters.getParameter("location")+"?"+request.getQueryString() + " caused an exception: " + e.getMessage());
			} catch (ParameterException e2) {
			}
		}
		for (int i = 0; i < locations.length; i++) {
			try {
				connection = getConnection(new URL(locations[i].getValue()+"?"+request.getQueryString()), username, password);
				return;
			} catch (Exception e) {
				try {
					System.out.println("HTTPServerTransformer at " + locations[i].getValue()+"?"+request.getQueryString() + " caused an exception: " + e.getMessage());
				} catch (ConfigurationException e2) {
				}
			}
		}
		throw new ProcessingException("All configured HTTPServers caused exceptions or no servers configured.");
	}
	
	/**
		*  Description of the Method
		*
		*@param  serverURL      Description of the Parameter
		*@param  queryString    Description of the Parameter
		*@param  username       Description of the Parameter
		*@param  password       Description of the Parameter
		*@return                Description of the Return Value
		*@exception  Exception  Description of the Exception
		*/
	   public void generate(String queryString) throws SAXException {
//		   System.out.println("Query: "+queryString);
	       connection.setDoOutput(true);
		   connection.setDoInput(true);
	   	   connection.setRequestProperty("Content-type", "text/xml");
		   connection.setRequestProperty("Content-Length", "" + queryString.length());
		   PrintWriter toServer = null;
		   try {
			   toServer = new PrintWriter(connection.getOutputStream());
		   } catch (IOException e) {
			   throw new SAXException("IOException outputting to connection: " + e.getMessage());
		   }
		   toServer.write(queryString);
		   toServer.flush();
		   toServer.close();

		   InputStream inputStream = null;
		   try {
			   inputStream = connection.getInputStream();
		   } catch (IOException e) {
			   throw new SAXException("IOException inputting from connection: " + e.getMessage());
		   }
		   ModifiableTimeStampValidity validity = (ModifiableTimeStampValidity)validityMap.get(getKey());		 
		   validity.setTimeStamp(connection.getLastModified());  
		   parse(inputStream);
	   }
	   
	public HttpURLConnection getConnection(URL serverURL, String username, String password) throws Exception {
		HttpURLConnection connection = (HttpURLConnection) serverURL.openConnection();
		if (username != null && password != null) {
			String toEncode = username + ":" + password;
			connection.setRequestProperty("Authorization",
					"Basic " + Base64.encodeBytes(toEncode.getBytes()));
		}
		return connection;
	}

	class ModifiableTimeStampValidity implements SourceValidity {
		private long timestamp=0;
		private boolean valid=false;
		
		public void setTimeStamp(long timestamp) {
			this.timestamp=timestamp;
//			System.out.println("New timestamp for "+/*getKey()+*/": "+timestamp);
			if (timestamp!=0) valid=true;
//			else valid=false;
		}
		
		public long getTimeStamp() {
			return this.timestamp; 
		}
		
		public int isValid() {
			if (valid) return 1;
			else return -1;
		}
		
		public int isValid(SourceValidity newValidity) {
			return isValid();
		}
	} 
}


